'use strict';

var PACKAGE_NAME = 'babel-plugin-partial-application';
var DEFAULT_PLACEHOLDER = '_';
var ALTERNATE_PLACEHOLDER = '__';
var POSITIONALS = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'];

function makeIdentifierError(path, current) {
  return path.buildCodeFrameError(`Cannot use placeholder as an identifier.\n
     If you need to use '${current}' as a variable name:
     - import or require the plugin to a new token
       ( \`import it from '${PACKAGE_NAME}'\` )
     - set 'options.placeholder' to a different custom token
     - set 'options.useAlternatePlaceholder' for '__'
       ( a double underscore )\n
    See https://github.com/citycide/${PACKAGE_NAME}#caveats--limitations
    for more information.\n`);
}

function makeDefaultArgError(path) {
  return path.buildCodeFrameError('Default parameters are only supported in argument lists.');
}

module.exports = function (_ref5) {
  var t = _ref5.types;

  var hash = new Map();

  function getPlaceholder(state) {
    var filename = state.file.opts.filename,
        useAlternatePlaceholder = state.opts.useAlternatePlaceholder;


    var placeholder = hash.get(filename) || state.opts.placeholder || DEFAULT_PLACEHOLDER;

    if (placeholder === DEFAULT_PLACEHOLDER && useAlternatePlaceholder) {
      placeholder = ALTERNATE_PLACEHOLDER;
    }var kind = typeof placeholder;
    if (kind !== 'string') {
      throw new TypeError('Expected a string for `options.placeholder` ' + `and got ${kind}`);
    }hash.set(filename, placeholder);
    return placeholder;
  }

  function setPlaceholder(state, token) {
    if (typeof token === 'string') {
      return hash.set(state.file.opts.filename, token);
    }
  }

  function getUniqueName(path) {
    var name = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'a';
    return path.scope.generateUidIdentifier(name);
  }

  function toNode(pathOrNode) {
    return (pathOrNode == null ? null : pathOrNode.node) || pathOrNode;
  }

  function isPlaceholder(target, placeholder) {
    return t.isIdentifier(toNode(target), { name: placeholder });
  }

  function isSpreadPlaceholder(target, placeholder) {
    var node = toNode(target);
    return t.isSpreadElement(node) && isPlaceholder(node.argument, placeholder);
  }

  function isPositionalPlaceholder(target, placeholder) {
    var _ref;

    var _toNode = toNode(target),
        tag = _toNode.tag,
        quasi = _toNode.quasi;

    return isPlaceholder(tag, placeholder) && POSITIONALS.indexOf(((_ref = quasi.quasis[0]) == null ? null : _ref.value) == null ? null : _ref.value.raw) > -1;
  }

  function hasPlaceholders(nodes, placeholder) {
    for (var _i6 = 0, _len7 = nodes.length; _i6 < _len7; _i6++) {
      var _target = nodes[_i6];

      var _node = toNode(_target);
      if (isPlaceholder(_node, placeholder) || isPositionalPlaceholder(_node, placeholder)) return true;
    }

    return false;
  }

  function isPartialCaller(target, placeholder) {
    var node = toNode(target);
    return t.isCallExpression(node) && hasPlaceholders(node.arguments, placeholder);
  }

  function isPluginRequire(node) {
    var callee = node.callee,
        args = node.arguments;


    var isRequire = t.isIdentifier(callee, { name: 'require' }) || t.isMemberExpression(callee) && t.isIdentifier(callee.object, { name: 'require' });

    return isRequire && t.isStringLiteral(args[0], { value: PACKAGE_NAME });
  }

  function pullPathLink(path) {
    if (path.isMemberExpression()) return path.get('object');else if (path.isCallExpression()) return path.get('callee.object');
  }

  function shouldHoist(target, placeholder) {
    var node = toNode(target);

    if (t.isTaggedTemplateExpression(node)) {
      return !isPositionalPlaceholder(node, placeholder);
    }if (t.isCallExpression(node)) {
      var _ref2;

      return !(((_ref2 = node.callee) == null ? null : _ref2.object) == null ? null : _ref2.object.wasPlaceholder);
    }if (t.isMemberExpression(node)) {
      return !node.object.wasPlaceholder;
    }return !t.isLiteral(node) && ['Identifier', 'SpreadElement', 'FunctionExpression', 'ArrowFunctionExpression'].every(function (type) {
      return type !== node.type;
    });
  }

  function makeWrapperFunction(params, returnValue) {
    var props = {
      positionals: { value: {} },
      isWrapper: { value: true }
    };

    return Object.defineProperties(t.arrowFunctionExpression(params, t.blockStatement([t.returnStatement(returnValue)])), props);
  }

  function findTopmostLink(path) {
    var tail = path;
    while (tail.parentPath) {
      var parent = tail.parentPath;
      var isCalleeTail = parent.isCallExpression() && parent.get('callee') === tail;
      var hasChainedParent = parent.isMemberExpression() || parent.isBinaryExpression() || isCalleeTail;

      if (!hasChainedParent) break;
      tail = tail.parentPath;
    }return tail;
  }

  function findDeepestLink(path) {
    var _ref6 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
        predicate = _ref6.predicate,
        fn = _ref6.fn;

    predicate = predicate || pullPathLink;
    fn = fn || predicate;

    var link = path;
    while (predicate(link)) {
      link = fn(link);
    }return link;
  }

  function findWrapperFunction(path) {
    return path.findParent(function (p) {
      return p.isArrowFunctionExpression() && p.node.isWrapper;
    });
  }

  function checkImports(path, state) {
    var placeholder = getPlaceholder(state);
    var _path$node = path.node,
        specifiers = _path$node.specifiers,
        source = _path$node.source;

    if (source.value === PACKAGE_NAME && specifiers.length) {
      setPlaceholder(state, specifiers[0].local.name);
      return path.remove();
    }
    for (var _i7 = 0, _len8 = specifiers.length; _i7 < _len8; _i7++) {
      var _specifier = specifiers[_i7];

      if (_specifier.local.name === placeholder) {
        throw makeIdentifierError(path, placeholder);
      }
    }
  }

  function checkDeclarations(path, state) {
    var placeholder = getPlaceholder(state);

    for (var _arr3 = path.node.declarations, _i8 = 0, _len9 = _arr3.length; _i8 < _len9; _i8++) {
      var _arr3$_i = _arr3[_i8],
          _init = _arr3$_i.init,
          _id = _arr3$_i.id;

      if (_init && isPluginRequire(_init) && t.isIdentifier(_id)) {
        setPlaceholder(state, _id.name);
        return path.remove();
      }if (_id.name === placeholder) {
        throw makeIdentifierError(path, placeholder);
      }
    }
  }

  function getPlaceholderPosition(path) {
    return Number(path.node.quasi.quasis[0].value.raw);
  }

  function parsePosition(node) {
    var fallback = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 99;

    var result = Number(node.position);
    return Number.isNaN(result) ? fallback : result;
  }

  function sortByPlaceholderPosition(a, b) {
    return parsePosition(a) > parsePosition(b);
  }

  var appliedChildVisitor = {
    CallExpression(childPath) {
      childPath.skip();
    }, Identifier(childPath) {
      var parent = childPath.parent,
          defaultParameter = childPath.node.defaultParameter;


      if (childPath.key === 'property' || !isPlaceholder(childPath, this.placeholder) || isPositionalPlaceholder(parent, this.placeholder)) return;

      var name = 'arg';
      if (t.isMemberExpression(parent)) {
        name = 'obj';
        parent.wasPlaceholder = true;
      }var id = getUniqueName(childPath, name);
      id.wasPlaceholder = true;
      if (defaultParameter) {
        this.fn.params.push(t.assignmentPattern(id, defaultParameter));
      } else {
        this.fn.params.push(id);
      }childPath.replaceWith(id);
    }, TaggedTemplateExpression(childPath) {
      if (!isPositionalPlaceholder(childPath.node, this.placeholder)) {
        return;
      }var position = getPlaceholderPosition(childPath);
      var current = this.fn.positionals[position];
      var id = current || getUniqueName(childPath, 'pos');
      id.position = position;

      if (!current) {
        this.fn.params.push(id);
        this.fn.positionals[position] = id;
      }this.fn.params.sort(sortByPlaceholderPosition);
      childPath.replaceWith(id);
      childPath.skip();
    }, SpreadElement(childPath) {
      var argument = childPath.node.argument;


      if (!isPlaceholder(argument, this.placeholder)) return;

      var id = getUniqueName(childPath, 'spr');
      this.fn.params.push(t.restElement(id));
      childPath.replaceWith(t.spreadElement(id));
    } };

  var binaryChildVisitor = {
    TaggedTemplateExpression(childPath) {
      if (!isPositionalPlaceholder(childPath.node, this.placeholder)) {
        return;
      }var position = getPlaceholderPosition(childPath);
      var current = this.tail.node.positionals[position];
      var id = current || getUniqueName(childPath, 'pos');
      id.position = position;

      if (!current) {
        this.tail.node.params.push(id);
        this.tail.node.positionals[position] = id;
      }this.tail.node.params.sort(sortByPlaceholderPosition);
      childPath.replaceWith(id);
      childPath.skip();
    }, Identifier(childPath) {
      if (!isPlaceholder(childPath.node, this.placeholder) || childPath.key === 'property') return;

      var id = getUniqueName(childPath, 'bin');
      this.tail.node.params.push(id);
      childPath.replaceWith(id);
    } };

  var templateLiteralVisitor = {
    TaggedTemplateExpression(childPath) {
      if (!isPositionalPlaceholder(childPath.node, this.placeholder)) {
        return;
      }var position = getPlaceholderPosition(childPath);
      var wrapper = findWrapperFunction(childPath);

      if (!wrapper) {
        var fn = makeWrapperFunction([], this.path.node);
        this.path.replaceWith(fn);
        // this is pretty hacky...
        wrapper = { node: fn };
      }var current = wrapper.node.positionals[position];
      var id = current || getUniqueName(childPath, 'pos');
      id.position = position;

      if (!current) {
        wrapper.node.params.push(id);
        wrapper.node.positionals[position] = id;
      }wrapper.node.params.sort(sortByPlaceholderPosition);
      childPath.replaceWith(id);
      childPath.skip();
    }, Identifier(childPath) {
      if (!isPlaceholder(childPath.node, this.placeholder)) {
        return;
      }if (childPath.key === 'property') return;

      var id = getUniqueName(childPath, 'exp');
      childPath.replaceWith(id);

      if (this.path.isArrowFunctionExpression()) {
        this.path.node.params.push(id);
        return;
      }if (this.path.container) {
        this.path.replaceWith(makeWrapperFunction([id], this.path.node));
      } else {
        var wrapper = findWrapperFunction(childPath);
        if (wrapper) wrapper.node.params.push(id);
      }this.path.stop();
    } };

  var assignmentVisitor = {
    AssignmentExpression(childPath) {
      var _childPath$node = childPath.node,
          left = _childPath$node.left,
          right = _childPath$node.right,
          listKey = childPath.listKey;


      if (isPlaceholder(left, this.placeholder)) {
        if (listKey !== 'arguments') {
          throw makeDefaultArgError(childPath);
        }left.defaultParameter = right;
        childPath.replaceWith(left);
        return;
      }var links = [];
      var link = left;
      while (link.object) {
        links.push(link.property);
        link = link.object;
      }if (isPlaceholder(link, this.placeholder)) {
        if (listKey !== 'arguments') {
          throw makeDefaultArgError(childPath);
        }link._meta = { links, right };
        childPath.replaceWith(left);
      }
    } };

  return {
    visitor: {
      Program(path, state) {
        var placeholder = getPlaceholder(state);
        path.traverse(assignmentVisitor, { placeholder });
      }, ImportDeclaration(path, state) {
        checkImports(path, state);
      }, VariableDeclaration(path, state) {
        checkDeclarations(path, state);
      }, CallExpression(path, state) {
        var placeholder = getPlaceholder(state);
        var args = path.get('arguments');

        if (!args.length || !hasPlaceholders(args, placeholder)) {
          return;
        }var fn = makeWrapperFunction([], path.node);
        path.traverse(appliedChildVisitor, { fn, placeholder });

        for (var _i9 = 0, _len10 = args.length; _i9 < _len10; _i9++) {
          var _arg = args[_i9];

          var hoist = shouldHoist(_arg, placeholder);
          if (hoist && !isPartialCaller(_arg, placeholder)) {
            var _id2 = getUniqueName(path, 'ref');
            var ref = t.variableDeclaration('const', [t.variableDeclarator(_id2, _arg.node)]);

            var parent = path.getStatementParent();
            parent.insertBefore(ref);
            _arg.replaceWith(_id2);
          }
        }

        path.replaceWith(fn);
      }, MemberExpression(path, state) {
        var _ref3, _ref4;

        var placeholder = getPlaceholder(state);
        var object = path.get('object');
        var node = object.node;


        if (!isPlaceholder(node, placeholder)) {
          return;
        }var last = ((_ref3 = node._meta) == null ? null : _ref3.links) == null ? null : _ref3.links[0];
        var param = (_ref4 = node._meta) == null ? null : _ref4.right;
        var pattern = void 0;
        if (last) {
          var name = getUniqueName(path, 'arg');
          pattern = t.objectPattern([t.objectProperty(last, t.assignmentExpression('=', name, param))]);

          if (node._meta.links.length > 1) {
            for (var _arr4 = node._meta.links, _i10 = 0, _len11 = _arr4.length; _i10 < _len11; _i10++) {
              var _link = _arr4[_i10];

              if (_i10 === 0) continue;
              pattern = t.objectPattern([t.objectProperty(_link, pattern)]);
            }
          }var top = findTopmostLink(path);
          if (top.parentPath.container) {
            top.parentPath.replaceWith(makeWrapperFunction([pattern], top.parent));
          } else {
            top.parent.params.push(pattern);
          }top.replaceWith(name);
          return;
        }var binaryParent = path.findParent(function (p) {
          return p.isBinaryExpression();
        });
        var ids = [getUniqueName(path, 'obj')];

        if (binaryParent) {
          var _right = binaryParent.get('right');
          if (!isPlaceholder(_right, placeholder)) {
            var _target2 = findDeepestLink(_right);
            if (isPlaceholder(_target2, placeholder)) {
              var _id3 = getUniqueName(path, 'obj');
              ids.push(_id3);
              _target2.replaceWith(_id3);
            }
          }
        }object.replaceWith(ids[0]);

        var tail = findTopmostLink(path);
        if (tail.container) {
          tail.replaceWith(makeWrapperFunction(ids, tail.node));
        } else {
          for (var _i11 = 0, _len12 = ids.length; _i11 < _len12; _i11++) {
            var _id4 = ids[_i11];

            tail.parent.params.push(_id4);
          }
        }
      }, SpreadElement(path, state) {
        var placeholder = getPlaceholder(state);
        var parent = path.parent,
            parentPath = path.parentPath;

        var argument = path.get('argument');

        if (!isPlaceholder(argument, placeholder)) return;

        if (path.listKey !== 'arguments') {
          throw path.parentPath.buildCodeFrameError('Placeholder spread is only supported as a function argument.');
        }
        function makeSpreadIdentifier() {
          return getUniqueName(path, 'spr');
        }

        var wrapper = findWrapperFunction(path);
        if (wrapper) {
          if (!path.container) return;

          var _id5 = makeSpreadIdentifier();
          wrapper.node.params.push(t.restElement(_id5));
          path.replaceWith(t.spreadElement(_id5));
        } else if (path.container) {
          var _id6 = makeSpreadIdentifier();
          path.replaceWith(t.spreadElement(_id6));
          parentPath.replaceWith(makeWrapperFunction([t.restElement(_id6)], parent));
        } else {
          var _id7 = makeSpreadIdentifier();
          parent.params.push(t.restElement(_id7));
          argument.replaceWith(_id7);
        }parentPath.stop();
      }, BinaryExpression(path, state) {
        var placeholder = getPlaceholder(state);

        var nodes = [path.get('left'), path.get('right')].filter(function (v) {
          return isPlaceholder(v, placeholder) || isPositionalPlaceholder(v, placeholder) || isPositionalPlaceholder(findDeepestLink(v), placeholder);
        });

        if (!nodes.length) return;

        var tail = path;
        while (tail.parentPath.isBinaryExpression()) {
          tail = tail.parentPath;
        }if (tail.container) {
          tail.replaceWith(makeWrapperFunction([], t.toExpression(tail.node)));
        } else {
          tail = tail.parentPath;
        }tail.traverse(binaryChildVisitor, { tail, placeholder });
        tail.stop();
      }, TemplateLiteral(path, state) {
        var placeholder = getPlaceholder(state);
        path.traverse(templateLiteralVisitor, { path, placeholder });
      }, TaggedTemplateExpression(path, state) {
        var parentPath = path.parentPath;

        var placeholder = getPlaceholder(state);
        if (!isPositionalPlaceholder(path, placeholder)) return;

        if (parentPath.isVariableDeclarator() || parentPath.isExpressionStatement()) {
          throw path.buildCodeFrameError('Positional placeholders must be part of some expression.');
        }var wasTransformed = t.isReturnStatement(parentPath.parent);

        var fn = void 0,
            current = void 0;
        if (wasTransformed) {
          fn = path.findParent(function (p) {
            return p.isArrowFunctionExpression();
          });
          current = fn.node.positionals[position];
        }var position = getPlaceholderPosition(path);
        var id = current || getUniqueName(path, 'pos');
        id.position = position;

        if (!wasTransformed) {
          var _target3 = findTopmostLink(path);
          if (_target3 === path) _target3 = parentPath;

          var wrapper = makeWrapperFunction([id], _target3.node);
          wrapper.positionals[position] = id;

          if (_target3.container) {
            _target3.replaceWith(wrapper);
          } else {
            _target3.parent.params.push(id);
            _target3.parent.params.sort(sortByPlaceholderPosition);
          }
        } else {
          if (!current) fn.node.params.push(id);
          fn.node.params.sort(sortByPlaceholderPosition);
        }path.replaceWith(id);
      } }
  };
};